function [SIGMA, ISIGMA, S] = GenerateData(varargin)
% GENERATEDATA Create a sparse precision matrix, its related covariance
% matrix and empirical covariance matrix.
% [SIGMA, ISIGMA, S, sigma] = GenerateData(varargin)
%
% Function which generates a sparse precision matrix ISIGMA at level P and
% dimension n, it computes at the same time the relative covariance matrix
% SIGMA. The position of non negative elements in ISIGMA are randomly
% selected via the native MatLab function RAND(), while N samples are drawn
% from a mvr Gaussian variable of zero mean and covariance SIGMA. The
% samples are corrupted by Gaussian noise of zero mean and variance
% N_SIGMA^2.
%
% The synopsis of the function is 
%
%   [SIGMA, ISIGMA, S] = GenerateData('Option1', Option1_value, ...
%                                      'Option2', Option2_value,...)
% 
% All the options provide a default value.
% This function is a slight modification of the code available in 
% http://stanford.edu/~boyd/papers/admm/covsel/covsel_example.html
%
% INPUT 
% 
% FEATS   : (integer) number of features. 
%                     Default: 100
% SAMPLES : (integer) number of samples to generate. 
%                     Default: 10*n.
% PERC    : (double)  percentage of nonzero elements in ISIGMA. 
%                     Default: 1e-3.
% NOISE   : (double)  noise level. 
%                     Default: 0;
% NSEED   : (integer) random seed for generating samples. 
%                     Default: 0.
% ESEED   : (integer) random seed for generating nonzero positions. 
%                     Default: 0.
%
% OUTPUTS
%
% SIGMA  : (double nxn array) Generated covariance matrix.
% ISIGMA : (double nxn array) Generated precision matrix.
% S      : (double nxn array) Empirical covariance matrix.
%
%
%==========================================================================
%
% Version : 1.0 (04-06-2019)
% Author  : A. Benfenati (alessandro.benfenati@unimi.it)
%
%==========================================================================
%
% COPYRIGHT NOTIFICATION
%
% Permission to copy and modify this software and its documentation for
% internal research use is granted, provided that this notice is retained
% thereon and on all copies or modifications. The authors and their
% respective Universities makes no representations as to the suitability
% and operability of this software for any purpose. It is provided "as is"
% without express or implied warranty. Use of this software for commercial
% purposes is expressly prohibited without contacting the authors.
%
% This program is free software; you can redistribute it and/or modify it
% under the terms of the GNU General Public License as published by the
% Free Software Foundation; either version 3 of the License, or (at your
% option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
% Public License for more details.
%
% You should have received a copy of the GNU General Public License along
% with this program; if not, either visite http://www.gnu.org/licenses/ or
% write to Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
% 02139, USA.
%


nseed   = 0;     %randn('seed', 0);
eseed   = 0;     %rand('seed', 0);
n       = 100;   % number of features
N       = 10*n;  % number of samples
p       = 0.001; % fraction of nonzero elements
n_sigma = 0;     % noise level


if (rem(length(varargin),2)==1)
    error('Optional parameters should always go by pairs');
else
    for i=1:2:(length(varargin)-1)
        switch upper(varargin{i})
            case 'FEATS'
                n = varargin{i+1};
            case 'SAMPLES'
                N = varargin{i+1};
            case 'PERC'
                p = varargin{i+1};
            case 'NOISE'
                n_sigma = varargin{i+1};
            case 'NSEED'
                nseed = varargin{i+1};
            case 'ESEED'
                eseed = varargin{i+1};
            otherwise
                error('Unknown Option.');
        end
    end
end


% Generate a sparse positive definite inverse covariance matrix
rand('seed',eseed);
ISIGMA      = diag(abs(ones(n,1)));
idx         = randsample(n^2, p*n^2);
ISIGMA(idx) = ones(numel(idx), 1);
ISIGMA      = ISIGMA + ISIGMA';     % make symmetric
if min(eig(ISIGMA)) < 0             % make positive definite
    ISIGMA = ISIGMA + 1.1*abs(min(eig(ISIGMA)))*eye(n);
end
SIGMA = inv(ISIGMA);

% generate Gaussian samples
randn('seed',nseed);
D = mvnrnd(zeros(1,n), SIGMA, N);

if n_sigma >0
    D = D + n_sigma*randn(N,n);
    S  = cov(D);
else
    S = cov(D);
end