function [Cn12, temp, fobj, Q,t_vec, ERR] = MMDR(Cl, gamma, mu0, mu1, S, sigma,varargin)
% MMDR Compute the estimation of precision matrix given the empirical
% coviariance matrix via the MM approach. It uses DR algorithm to solve
% inner optimization problems
%
%   [Cn12, temp, fobj, Q,t_vec, ERR] = MMDR(Cl, gamma, mu0, mu1, S,
%   sigma,varargin)
% 
% It solves the problem
%
% argmin logdet(C^{-1} + s^2*Id) + trace( C*S*(Id + s^2 C)^(-1)) +
%        + mu_0*g_0(C) + mu_1*g_1(C)
%
% adopting an MM approach. The inner subproblems arising from this approach
% are solve via the DR algorithm. See Benfenati et al, "Proximal Approaches
% for Matrix Optimization Problems: Application to Robust Precision
% MatrixEstimation." for technical details.
%
% -------------------------------------------------------------------------
% MANDATORY INPUT
% Ck        : (double array) initial estimate
% gamma     : (double)       parameter for the proximal operator
% mu0       : (double)       regularization parameter for g0
% mu1       : (double)       regularization parameter for g1
% S         : (double array) given symmetric semidefinite positive matrix
% sigma     : (double)       noise level
%
% -------------------------------------------------------------------------
% OPTIONAL INPUT
%
% INN_ITER  : (integer)      maximum number of DR iterations.
%                            Default: 100;
% INN_STOP  : (integer)      stopping criterion for DR. Option 1 stands for
%                            running all the iterations; options 2 stands
%                            for controlling the relative decreasing of the
%                            objective function value below the tolerance
%                            inn_tol. Options 3 stands for checking the
%                            relative distance between two successive
%                            iterations. Default: 1;
% INN_TOL   : (double)       tolerance for the inner stopping criterion.
%                            Default: 1e-6;
% EXT_ITER  : (integer)      maximum number of MM iterations.
%                            Default: 1000;
% EXT_STOP  : (integer)      stopping criterion for MM. Option 1 stands for
%                            running all the iterations; options 2 stands
%                            for controlling the relative decreasing of the
%                            objective function value below the tolerance
%                            inn_tol. Default: 1;
% EXT_TOL   : (double)       tolerance for the external stopping criterion.
%                            Default: 1e-6;
% LAMBDA    : (double)       DR parameter.
%                            Default: 1.5.
% F         : (string)       Fit-to-data functional.
%                            Default: "LogDet".
% G0        : (string)       eigenvalues regularization functional,
%                            Default: Shatten--1 norm.
% G1        : (string)       regularization functional,
%                            Default: l_1 norm.
% OBJ       : (double array) Benchmark for computing performances.
%                            Default: empty.
% PARAMS    : (double array) parameters needed for computing g0 and its
%                            proximal operators. See the help of the
%                            related function
% VERBOSE   : (integer)      Verbosity. Option 1 (or greater than 0) prints
%                            iterations 'information on the shell. Option 0
%                            stands for silence. Default: 1.
%
% -------------------------------------------------------------------------
% OUTPUT
%
% Cn12      : (double array) estimated solution
% temp      : (double array) estimated solution (At convergence, Ck=Ck12)
% fobj      : (double array) vector of functional's values
% Q         : (cell)         cell containing the objective functional related to each subproblem.
% t_vec     : (double array) employed time per iteration
% ERR       : (double array) if obj is not empty, relative reconstruction error wrt obj
%
%
%==========================================================================
%
% Version : 1.0 (04-06-2019)
% Author  : A. Benfenati (alessandro.benfenati@unimi.it)
%
%==========================================================================
%
% COPYRIGHT NOTIFICATION
%
% Permission to copy and modify this software and its documentation for
% internal research use is granted, provided that this notice is retained
% thereon and on all copies or modifications. The authors and their
% respective Universities makes no representations as to the suitability
% and operability of this software for any purpose. It is provided "as is"
% without express or implied warranty. Use of this software for commercial
% purposes is expressly prohibited without contacting the authors.
%
% This program is free software; you can redistribute it and/or modify it
% under the terms of the GNU General Public License as published by the
% Free Software Foundation; either version 3 of the License, or (at your
% option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
% Public License for more details.
%
% You should have received a copy of the GNU General Public License along
% with this program; if not, either visite http://www.gnu.org/licenses/ or
% write to Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
% 02139, USA.
%

% Default options
INN_ITER    = 100;
inn_stop    = 1;
inn_tol     = 1e-6;
EXT_ITER    = 1000;
ext_stop    = 1e-6;
ext_tol     = 1e-6;
lambda      = 1.5;
FitData     = 'LogDet';
g0          = 'iSchatten1';
g1          = 'ell1';
verbose     = 1;
params      = [mu0, sigma];
obj         = [];

if (nargin-length(varargin)) ~= 6
    error('Wrong number of required parameters');
end

if (rem(length(varargin),2)==1)
    error('Optional parameters should always go by pairs');
else
    for i=1:2:(length(varargin)-1)
        switch upper(varargin{i})
            case 'INNITER'
                INN_ITER = varargin{i+1};
            case 'INNSTOP'
                inn_stop = varargin{i+1};
            case 'INNTOL'
                inn_tol = varargin{i+1};
            case 'EXTITER'
                EXT_ITER = varargin{i+1};
            case 'EXTSTOP'
                ext_stop = varargin{i+1};
            case 'EXTTOL'
                ext_tol = varargin{i+1};
            case 'LAMBDA'
                lambda = varargin{i+1};
            case 'F'
                FitData = varargin{i+1};
            case 'G0'
                g0 = varargin{i+1};
            case 'G1'
                g1 = varargin{i+1};
            case 'VERBOSE'
                verbose = varargin{i+1};
            case 'OBJ'
                obj = varargin{i+1};
            case 'PARAMS'
                params = varargin{i+1};
                
        end
    end
end

% Function Handles
prox_g0  = str2func(sprintf('@prox_%s_%s',FitData,g0));
prox_g1  = str2func(sprintf('@prox_%s',g1));
Fit2Data = str2func(FitData);
g0       = str2func(g0);
g1       = str2func(g1);

% First estimation
% Computations to be done once
Id      = eye(size(Cl));

% Allocating memory
fobj        = zeros(INN_ITER+1,1);
Q           = cell(EXT_ITER,1);
t_vec       = zeros(INN_ITER+1,1);
ERR         = zeros(INN_ITER+1,1);


if ~isempty(obj)
    norm_obj    = norm(obj,'fro');
    ERR(1)      = norm(Cl-obj,'fro')/norm_obj;
end

t_vec(1)    = 0;


l        = 1;
ext_loop = 1;

fobj(1) = Fit2Data((Id+sigma^2*Cl)\Cl,S) + mu0*g0(Cl,params) + mu1*g1(Cl);

if verbose
    fprintf('%5d) Fobj: %+4.8f ',0,fobj(1));
    if ~isempty(obj)
        fprintf('RMSE : %2.2f',ERR(1));
    end
    fprintf('\n')
end
% [~,dc]  = mexeig(Cl);
% dc      = diag(dc);
dc = eig(Cl);


tic;
temp = 0;
while ext_loop
    
    %% Computations needed once per external iteration
    grad        = ((Id+sigma^2*Cl)\S) / (Id+sigma^2*Cl); % checked
    trCl        = trace(((Id+sigma^2*Cl)\Cl)*S);         % checked
    grad_TrCl   = trace(grad*Cl);                        % checked
    
    
    Q{l}    = zeros(INN_ITER+1,1);
    Q{l}(1) = -sum(log(u(dc,sigma))) + trCl + mu0*g0(dc,params) + mu1*g1(Cl);
    n       = 1;
    
    %% DOUGLAS RACHFORD
    inn_loop = 1;
    Cn = Cl;
    nCL = norm(Cl,'fro');
    while inn_loop
        % First operator
        Cn12    = prox_g1(Cn,gamma,mu1);
%         [~,dc]  = mexeig(Cn12);
%         dc      = diag(dc);
        dc = eig(Cn12);        
        
        % Computing the Objective
        Q{l}(n+1) = -sum(log(u(dc,sigma))) +...
            trCl + trace(grad*Cn12) - grad_TrCl +...
            mu0*g0(dc,params)+ mu1*g1(Cn12);
        
        if verbose
            fprintf('%5d--%5d) Q: %+4.8f',l,n,Q{l}(n+1));
            fprintf('\n')
        end
        switch inn_stop
            case 1
                % Number of iterations
                if n>INN_ITER-1
                    inn_loop = 0;
                    Q{l}(n+2:end) = [];
                end
            case 2
                % Relative decrease of the objective
                if (((abs(Q{l}(n+1)-Q{l}(n))/abs(Q{l}(n))<inn_tol)) || (n>INN_ITER-1))
                    inn_loop = 0;
                    Q{l}(n+2:end) = [];
                end
            case 3
                % Difference between the two DR points
                if (n>1) && ((norm(temp-Cn12,'fro')<inn_tol ) || (n>INN_ITER-1))
                    inn_loop = 0;
                    Q{l}(n+2:end) = [];
                end
            otherwise  % Maybe this check has to be done at the beginning
                error('Invalid criterion.');
        end
        
        if inn_loop
            n = n+1;
            % Second operator
            temp = 2*Cn12-Cn;
            temp = prox_svd(temp - grad, gamma, prox_g0, params);
            
            % Final update
            Cn = Cn + lambda * (temp - Cn12);
            t_vec(n) = toc;
        end
        
        
    end
    %% End of DR
    
    fobj(l+1) = Fit2Data((Id+sigma^2*Cn12)\Cn12,S,u(dc,sigma)) + mu0*g0(dc,params) + mu1*g1(Cn12);
    
    if ~isempty(obj)
        ERR(l+1)      = norm(Cl-obj,'fro')/norm_obj;
    end
    old = Cl;
    Cl = Cn12;
    
    switch ext_stop
        case 1
            if l>EXT_ITER-1
                ext_loop = 0;
            end
        case 2
            if (((abs(fobj(l+1)-fobj(l))/abs(fobj(l))<ext_tol)) || (l>EXT_ITER-1)) && (l>2) 
                ext_loop = 0;
                Q{l}(n+2:end) = [];
            end
        case 3 
            if ( (norm(Cn12-old,'fro')/nCL<ext_tol) || (l>EXT_ITER-1))  
                ext_loop = 0;
                Q{l}(n+2:end) = [];
            end
    end
    l = l+1;
end

% Cutting off the extra elements
fobj(l+1:end)    = [];
t_vec(l+1:end)   = [];
if ~isempty(obj)
    ERR(l+1:end)    = [];
end
end

% See Lemma 1 of Benfenati et al, "Proximal Approaches for Matrix
% Optimization Problems: Application to Robust Precision MatrixEstimation."
function U = u(x,sigma)

U = x./(1+sigma^2*x);

end

